#!/usr/bin/perl -w

#   Copyright (C) 2006 David G. Jacobowitz 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

#perl2exe_include "Statistics/Distributions.pm"

use fgroup;
package fgroup::anova;

BEGIN { @ISA = qw(fgroup); };

use Exporter 'import';

@EXPORT_OK = qw();

use misc qw(fixedFloat numCount min);
use dataqual qw(gen_defquald_ha);

use misc2 qw(func_fail inDict inAry);

use Statistics::Distributions qw(chisqrprob fdistr fprob);
use Data::Dumper;
use eqn_string qw(create_evaluable_eqn);

sub commands {
 my $self = shift;
 my @cmds = ("oneway", "anova");
 return @cmds;
}

sub oneway {
 my $self = shift;
 my $a    = shift;
 my $r = $self->anova($a,"oneway");
 return $r;
}

sub anova {
 my $self   = shift;
 my $a      = shift;
 my $oneway = shift;
 my $argsr  = $a->{args};
 my $flgsr  = $a->{flgs};


 if (!defined($oneway)) { $oneway = "anova"; };
 
 my @vars;
 foreach my $pv (@{$a->{varary}}) {
  if (defined($pv->{state}) && ($pv->{state} eq "exist")) {
   push(@vars, $pv->{name});
  }
 }

 my $var1  = shift @vars;
 my $var2  = shift @vars;

 if (!defined($var1)) {
  return { status => "fail; no variable provided" };
 }
 if (!defined($var2)) {
  return { status => "fail; second variable not provided" };
 }
 if (!inDict($var1,$self->{dm}{data})) {
  if (!defined($var1)) { $var1 = ""; };
  return { status => "fail; var $var1 unknown" };
 }
 if (!inDict($var2,$self->{dm}{data})) {
  return { status => "fail; var $var1 unknown" };
 };
 if (@vars) {
  return { status => "fail; too many vars; only 1-way supported right now" }
 }

 my $r = { status => "ok" };


 @vars = ($var1, $var2);
 my $qdata = gen_defquald_ha($self->{dm},\@vars,$a->{ifary});

 if ($qdata->{info}{$var1}{type} ne "numeric") {
  return { status => "fail; var $var1 not wholly numeric" };
 }

 my $bybins  = {};
 my $bybinct = {};
 my @bydata  = @{$qdata->{data}{$var2}};
 for (my $row=0;$row<@bydata;$row++) {
  if (!($r->{have_if}) || $r->{ifary}[$row]) {
   if (!defined($bybinct->{$bydata[$row]})) {
    $bybinct->{$bydata[$row]} = 0;
   } else {
    $bybinct->{$bydata[$row]} += 1
   }
   $bybins->{$bydata[$row]}[$bybinct->{$bydata[$row]}] =
     $qdata->{data}{$var1}[$row];
  }
 }
 my $aa = [];
 my $ct = 0;
 foreach $k (keys %{$bybins}) {
  $aa->[$ct++] = $bybins->{$k}
 }

 $r = $self->calc_oneway_anova($aa);
 $r->{byvar} = $var2;
 if (func_fail($r)) { return $r; };
 if ($oneway eq "oneway") {
  $r = $self->oneway_prettyprint($r);
 } else {
  $r = $self->anova_prettyprint($r);
 }
 return $r;
}


sub anova_a {
 my $self  = shift;
 my $argsr = shift;
 my @vars  = @{$argsr};
 foreach my $v (@vars) {
  # print "v $v\n";
  if (!inDict($v,$self->{dm}{data})) {
   return { status => "fail; variable $v unknown" };
  }
 }
 my $aa    = [];
 my $count = 0;
 foreach my $v (@vars) {
  $aa->[$count++] = $self->{dm}{data}{$v};
 };

 my $r = $self->calc_oneway_anova($aa);
 return $r;
}





sub calc_oneway_anova {

 my $self  = shift;
 my $data  = shift;
 my $r     = {};

 my @cat_ary_refs = @{$data};

 $r->{k}          = $#cat_ary_refs + 1;

 $r->{total} = 0;
 $r->{N}     = 0;
 my $cat = 0;	
 foreach my $cat_ary_ref (@cat_ary_refs) {
  my @cat_ary = @{$cat_ary_ref};
  if (@cat_ary < 2) {
   return { status => "fail; too few elements in categories. Must be >=2" }
  }
  my $cat_total = 0;
  for (my $i=0;$i<@cat_ary;$i++) { $cat_total += $cat_ary[$i]; }
  my $cat_mean = $cat_total / @cat_ary;
  my $ss = 0;
  for (my $i=0;$i<@cat_ary;$i++) {
   my $diff = $cat_ary[$i] - $cat_mean;
   my $sq   = $diff*$diff;
   $ss += $sq; 
  }
  $r->{cats}[$cat]{sum_squares} = $ss;
  $r->{cats}[$cat]{n}           = @cat_ary;
  $r->{cats}[$cat]{mean}        = $cat_mean;
  $r->{cats}[$cat]{total}       = $cat_total;
  $r->{cats}[$cat]{variance}    = $ss / (@cat_ary - 1);
  $r->{total}                  += $cat_total;
  $r->{N}                      += @cat_ary;
  $cat++;
 }

 $r->{mean}  = $r->{total} / $r->{N};

 my $SSB = 0;
 my $SSW = 0;
 my $T_num_right = 0;
 my $T_den_a = 0;
 my $cats_skipped = 0;
 my $obs_skipped  = 0;
 for (my $c=0;$c<@{$r->{cats}};$c++) {
  $SSB += $r->{cats}[$c]{n} * ($r->{cats}[$c]{mean} - $r->{mean})**2;
  $SSW += ($r->{cats}[$c]{n} - 1)*($r->{cats}[$c]{variance});
  if ($r->{cats}[$c]{variance} > 0) {
   $T_num_right += ($r->{cats}[$c]{n} - 1) * log($r->{cats}[$c]{variance});
   $T_den_a += 1 / ($r->{cats}[$c]{n} - 1);
  } else {
   $cats_skipped += 1;
   $obs_skipped += $r->{cats}[$c]{n};
  }
 }
 $r->{vari_p}    = $SSW / ($r->{N} - $r->{k});
 $r->{SSB}       = $SSB;
 $r->{dof_numer} = ($r->{k} - 1);
 $r->{MSB}       = $SSB / $r->{dof_numer};
 $r->{SSW}       = $SSW;
 $r->{dof_denom} = $r->{N} - $r->{k};
 $r->{MSW}       = $SSW / $r->{dof_denom};
 $r->{SST}       = $SSB + $SSW;
 $r->{MST}       = $r->{SST} / ($r->{N} - 1);
 $r->{root_MSE}  = sqrt($r->{MSW});
 $r->{rsquared}  = $r->{SSB} / $r->{SST};
 #$r->{rsquared}  = $r->{MSB} / $r->{SSB};
 $r->{adj_rsquared} = 999.999;
 $r->{F}      = $r->{MSB}/$r->{MSW};
 $r->{F_crit} = fdistr($r->{dof_numer},
	               $r->{dof_denom},
	               (1-$self->{gb}{conf})/2);
 $r->{F_prob} = fprob($r->{dof_numer},
	              $r->{dof_denom},
                      $r->{F});
 $r->{var_p} = $SSW / ($r->{N} - $r->{k});

 ## calculate Bartlett's watchahoosie
 if ($cats_skipped) {
  $r->{N} -= $obs_skipped;
  $r->{k} -= $cats_skipped;
 }

 my $T_num = ($r->{N} - $r->{k})*log($r->{var_p}) - $T_num_right;
 my $T_den = 1 + (1/(3*($r->{k} - 1)))*(($T_den_a) - (1/ ($r->{N} - $r->{k})));
 my $T_bartlett = $T_num / $T_den;
 if ($cats_skipped) { $r->{bartlett_err} = 1; } else { $r->{bartlett_err} = 0; };

 $r->{chisq_bartlett} = $T_bartlett;
 $r->{P_chisq_bartlett} = chisqrprob($r->{dof_numer},$r->{chisq_bartlett});
 $r->{status} = "ok";
 return $r;
}


sub anova_prettyprint {
 my $self = shift;
 my $r    = shift;

 my $os   = "\n\n";
 $os .= sprintf("           Number of obs = %7d     R-squared     = %s\n",$r->{N},fixedFloat($r->{rsquared}));
 $os .= sprintf("           Root MSE      = %s  Adj R-squared = %s\n",fixedFloat($r->{root_MSE}),1?"_n/a_":fixedFloat($r->{adj_rsquared})); 
 $os .= "\n";
 $os .= "        Source |   Partial SS     df         MS          F   Prob > F\n";
 $os .= "    -----------+-----------------------------------------------------\n";
 $os .= sprintf("         Model |   %s %6d %s %s %s\n",fixedFloat($r->{SSB}),fixedFloat($r->{dof_numer}),fixedFloat($r->{MSB}),fixedFloat($r->{F}),fixedFloat($r->{F_prob}));
 $os .= "               |\n";
 $os .= sprintf(" %13s |   %s %6d %s %s %s\n",$r->{byvar},fixedFloat($r->{SSB}),fixedFloat($r->{dof_numer}),fixedFloat($r->{MSB}),fixedFloat($r->{F}),fixedFloat($r->{F_prob}));
 $os .= "               |\n";
 $os .= sprintf("      Residual |   %s %6d %s\n",fixedFloat($r->{SSW}),fixedFloat($r->{dof_denom}),fixedFloat($r->{MSW}));
 $os .= "    -----------+-----------------------------------------------------\n";
 $os .= sprintf("         Total |   %s %6d %s\n",fixedFloat($r->{SST}),$r->{N} -1,fixedFloat($r->{MST}));
 $os .= "\n\n";
 $os .= "\n";
 my $oo = $self->{oo};
 $oo->disp($os);
 return $r;
}


sub oneway_prettyprint {
 my $self = shift;
 my $r    = shift;
 my $os   = "\n\n";
 $os .= "                        Analysis of Variance\n";
 $os .= "    Source                  SS     df         MS          F   Prob > F\n";
 $os .= "------------------------------------------------------------------------\n";
 $os .= sprintf("Between groups      %s %6d %s %s %s\n",
	 fixedFloat($r->{SSB}),
	 $r->{dof_numer},
	 fixedFloat($r->{MSB}),
         fixedFloat($r->{F}),
         fixedFloat($r->{F_prob}));
 $os .= sprintf("Within groups       %s %6d %s\n",
         fixedFloat($r->{SSW}),
	 $r->{dof_denom},
	 fixedFloat($r->{MSW}));
 $os .= "------------------------------------------------------------------------\n";
 $os .= sprintf("Total               %s %6d %s\n",
         fixedFloat($r->{SST}),
	 $r->{N} - 1,
	 fixedFloat($r->{MST}));
 $os .= "\n";
 $os .= "Bartlett's test for equal variances:\n";
 $os .= sprintf("    chi2(%d) = %s Prob>chi2 = %s\n",$r->{dof_numer},fixedFloat($r->{chisq_bartlett}),fixedFloat($r->{P_chisq_bartlett}));
 if ($r->{bartless_err}) {
  $os .= "\nNote: Bartlett's test inaccurate; some category bins had zero variance.\n";
 }
 $os .= "\n\n";
 my $oo = $self->{oo};
 $oo->disp($os);
 return $r;
}


1;

