#!/usr/bin/perl -w

#   Copyright (C) 2006 David G. Jacobowitz 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

use fgroup;
package fgroup::tab;
BEGIN { @ISA = qw(fgroup); };

use Exporter 'import';
@EXPORT_OK = qw();

use warnings;

use strict;
use misc qw(fixedFloat);
use dataqual qw(gen_defquald_aa gen_defquald_ha gen_defquald_a
                gen_ifquald_aa  gen_ifquald_ha  gen_ifquald_a);
use misc2 qw(inDict inAry tstFlg func_fail);
use Scalar::Util qw(looks_like_number);
use gui qw(tk_updateVars);
use Data::Dumper;

sub commands {
 my $self = shift;
 my @cmds = ("tab", "csgof");
 return @cmds;
}


sub clip7 {
 my $is = shift;
 $is .= "                                    ";
 my $os = substr($is,0,7);
 return $os;
}

sub chi2_twoway {
 my $self    = shift;
 my $xya     = shift;

 my @ct = ();
 my @rt = ();
 my $x  = 0;
 my $y  = 0;
 my $v  = 0;

 my $cols = @{$xya};
 my $rows = @{$xya->[0]};

 if ($rows < 2) {
  return { status => "fail; insufficient rows in table for chi2" }
 }
 if ($cols < 2) {
  return { status => "fail; insufficient cols in table for chi2" }
 }

 for (my $y=0;$y<$rows;$y++) {
  $rt[$y] = 0;
 }

 for (my $x=0;$x<$cols;$x++) {
  $ct[$x] = 0;
  for (my $y=0;$y<$rows;$y++) {
   $v = $xya->[$x][$y];
   # print "X $x Y $y V $v\n";
   $ct[$x] += $v;
   $rt[$y] += $v;
  }
 }

 my $tt = 0;
 for (my $y=0;$y<$rows;$y++) {
  $tt += $rt[$y];
 }

 my $ott = 0;
 for (my $x=0;$x<$cols;$x++) {
  $ott += $ct[$x];
 }

 my $xyoe2e = [];
 my $chi2   = 0;
 for (my $x=0;$x<$cols;$x++) {
  for (my $y=0;$y<$rows;$y++) {
   # print "X $x Y $y COLS $cols ROWS $rows CT $ct[$x] RT $rt[$y] TT $tt\n";
   my $e = ($ct[$x] * $rt[$y]) / $tt;
   my $o = $xya->[$x][$y];
   my $oe2e = (($o - $e)*($o - $e))/$e;
   $xyoe2e->[$x][$y] = $oe2e; 
   $chi2 += $oe2e;
  }
 }

 my $r = {};
 $r->{chi2}   = $chi2;
 my $dof      = ($rows-1)*($cols-1);
 $r->{df}     = $dof;
 my $p_chi2   = Statistics::Distributions::chisqrprob($dof,$chi2);
 $r->{P_chi2} = $p_chi2;
 
 my $oo = $self->{oo};
 $oo->disp("\n       Pearson chi2($dof) = $chi2 Pr=$p_chi2\n");
 $r->{status} = "ok";
 return $r;
}

sub tab_twoway {
 my $self    = shift;
 my $xn      = shift;
 my $yn      = shift;
 my $a       = shift;

 # this is the maximum number of columns we can fit in an eighty-char 
 # wide screen
 my $max_cols_80 = 7;

 my $dochi2  = tstFlg($a,"chi2");
 my $force   = tstFlg($a,"force");
 my $missing = tstFlg($a,"missing");
 my $showrow = tstFlg($a,"row");

 my $r       = { status => "ok" };

 my $oo      = $self->{oo};

 my $tr = undef;

 my @varlist = ($xn, $yn);
 my $qdm;
 if ($missing) {
  $qdm = gen_ifquald_ha($self->{dm},\@varlist,$a->{ifary});
 } else {
  $qdm = gen_defquald_ha($self->{dm},\@varlist,$a->{ifary});
 }
 my %xh;
 my %yh;

 # print Dumper($qdm);

 if (!defined($qdm->{data}{$xn})) {
  return { status => "fail; empty array in $xn" };
 }

 for (my $i=0;$i<@{$qdm->{data}{$xn}};$i++) {
  if (!$a->{if_ok} || $a->{ifary}[$i]) {
   if (defined($xh{$qdm->{data}{$xn}[$i]})) {
    $xh{$qdm->{data}{$xn}[$i]} += 1;
   } else {
    $xh{$qdm->{data}{$xn}[$i]} = 1;
   }
   if (defined($yh{$qdm->{data}{$yn}[$i]})) {
    $yh{$qdm->{data}{$yn}[$i]} += 1;
   } else {
    $yh{$qdm->{data}{$yn}[$i]} = 1;
   }
  }
 }

 my %xyh;
 my $xya = [];
 my $y=0;
 foreach my $ym (sort keys %yh) {
  my $x=0;
  foreach my $xm (sort keys %xh) {
   $xyh{$xm}{$ym} = 0;
   $xya->[$x][$y] = 0;
   for (my $i=0;$i<@{$qdm->{data}{$xn}};$i++) {
    if (!$a->{if_ok} || $a->{ifary}[$i]) {
     if (($qdm->{data}{$xn}[$i] eq $xm) &&
         ($qdm->{data}{$yn}[$i] eq $ym)) {
      $xyh{$xm}{$ym} += 1;
      $xya->[$x][$y] += 1;
     }
    }
   }
   $x++;
  }
  $y++;
 }

 my @xk = sort keys %xh;
 my $xc = @xk;
 if (!$force && ($xc > $max_cols_80)) {
  return { status => "fail; across var has too many categories to display" }
 }

 my $os = "\n\n";
 $os .= sprintf("%s | %s\n",clip7(""),clip7($xn));
 $os .= sprintf("%s | ",clip7(""));
 foreach my $xm (sort keys %xh) {
  $os .= clip7($xm) . " ";
 }
 $os .= "  (total) ";
 $os .= "\n"; 
 my $dashes = "---------------";

 $os .= sprintf("%s + ",clip7($dashes));
 for (my $i=0;$i<$xc;$i++) {
  $os .= clip7($dashes) . " ";
 }
 $os .= "| " . clip7($dashes) . "\n";

 $os .= sprintf("%s | ",clip7($yn));
 for (my $i=0;$i<$xc;$i++) {
  $os .= clip7("") . " ";
 }
 $os .= "|\n";

 $os .= sprintf("%s | ",clip7(""));
 for (my $i=0;$i<$xc;$i++) {
  $os .= clip7("") . " ";
 }
 $os .= "|\n";

 my %ct;
 foreach my $xm (sort keys %xh) { $ct{$xm} = 0; };
 foreach my $ym (sort keys %yh) {
  $os .= sprintf("%s | ",clip7($ym));
  my $rt = 0;
  foreach my $xm (sort keys %xh) {
   my $v= $xyh{$xm}{$ym};
   $rt += $v;
   $ct{$xm} += $v;
   $os .= clip7($v) . " ";
  }
  $os .= "| " . clip7($rt);
  $os .= "\n";
  if ($showrow) {
   $os .= sprintf("%s | ",clip7(""));
   foreach my $xm (sort keys %xh) {
    my $v= $xyh{$xm}{$ym};
    my $pct = int(($v/$rt)*10000)/100;
    $os .= clip7($pct . "%") . " ";
   }
   $os .= "| 100%";
   $os .= "\n";
  }
 }

 $os .= sprintf("%s + ",clip7($dashes));
 for (my $i=0;$i<$xc;$i++) {
  $os .= clip7($dashes) . " ";
 }
 $os .= "| " . clip7($dashes) . "\n";

 $os .= sprintf("%s | ",clip7("")); #$ym
 my $tt = 0;
 foreach my $xm (sort keys %xh) {
  my $tv = $ct{$xm};
  $tt += $tv;
  $os .= clip7($tv) . " ";
 };
 $os .= "| " . clip7($tt);
 $os .= "\n";
 if ($showrow) {
  $os .= sprintf("%s | ",clip7("")); #$ym
  my $tt = 0;
  foreach my $xm (sort keys %xh) {
   my $tv = $ct{$xm};
   $tt += $tv;
  };
  foreach my $xm (sort keys %xh) {
   my $tv = $ct{$xm};
   my $pct = int(($tv/$tt)*10000)/100;
   $os .= clip7($pct . "%") . " ";
  };
  $os .= "| 100%";
  $os .= "\n";
 }
 $oo->disp($os);
 $r->{xy_tab} = \%xyh;


 if ($dochi2) {
  $self->chi2_twoway($xya);
 }
 $oo->disp("\n");
 return $r;
}


sub tabgen {
 my $dm      = shift;
 my $origvar = shift;
 my $newbase = shift;
 my $val     = shift;

 my $newvar  = $newbase . "_" . $val;

 for (my $i=0;$i<@{$dm->{data}{$origvar}};$i++) {
  if (looks_like_number($val) &&
      looks_like_number($dm->{data}{$origvar}[$i])) {
   $dm->{data}{$newvar}[$i] = ($val == $dm->{data}{$origvar}[$i]) ? 1 : 0;
  } else {
   $dm->{data}{$newvar}[$i] = ($val eq $dm->{data}{$origvar}[$i]) ? 1 : 0;
  }
  $dm->{info}{$newvar}{type} = "numeric";
 }
}



sub tab {
 my $self    = shift;
 my $ar      = shift;
 my $argref  = $ar->{args};
 my $flagref = $ar->{flgs};
 my $oo      = $self->{oo};
 my $r = { status => "ok" };
 my %histo;

 my @vars = ();
 foreach my $v (@{$ar->{varary}}) {
  push (@vars, $v->{name}) unless ($v->{state} ne "exist");
 }

 if (@vars == 2) {
  if (($ar->{varary}[0]{state} eq "exist") &&
      ($ar->{varary}[1]{state} eq "exist")) {
   my $xv = $ar->{varary}[0]{name};
   my $yv = $ar->{varary}[1]{name};
   $r = $self->tab_twoway($xv,$yv,$ar);
  } else {
   return { status => "fail; both vars not found" }
  }
  return $r;
 } elsif (@vars == 1) {
  if (inDict($vars[0], $self->{dm}{data})) {
   my $var = $vars[0];
   my $ary_r;
   my $missing = tstFlg($ar,"missing");
   if ($missing) {
    $ary_r = gen_ifquald_a($self->{dm},$var,$ar->{ifary});
   } else {
    $ary_r = gen_defquald_a($self->{dm},$var,$ar->{ifary});
   }

   if (!defined($ary_r)) {
    return { status => "fail; empty array for variable; no observations?" };
   }

   my @ary = @{$ary_r};
   my $qualct = 0;
   for (my $i=0;$i<@ary;$i++) {
    if (defined($histo{$ary[$i]})) {
     $histo{$ary[$i]} += 1;
    }  else {
     $histo{$ary[$i]} = 1;
    }
    $qualct++;
   }
   $oo->disp("\n");
   $oo->disp(sprintf("%10s | %10s | %10s | %10s\n",$vars[0],"Freq.","Percent","Cumulative"));
   my $dashes = "----------";
   $oo->disp(sprintf("%10s-+-%10s-+-%10s-+-%10s\n",$dashes,$dashes,$dashes,$dashes,$dashes));
   my $cum_percent = 0;
   foreach my $k (sort { (looks_like_number($a) && looks_like_number($b)) ? ($a <=> $b) : ($a cmp $b) } keys %histo) {
    my $curr_percent = 100 * $histo{$k} / $qualct;
    $cum_percent  += $curr_percent;
    $oo->disp(sprintf("%10.10s | %10s | %10f | %10f\n",$k,$histo{$k},
	                  $curr_percent, $cum_percent));
    if ($ar->{have_gen}) {
     tabgen($self->{dm},$vars[0],$ar->{genvar},$k);
    }
   }
   if ($ar->{have_gen}) {
    tk_updateVars($self->{dm});
   }
   $oo->disp(sprintf("%10.10s-+-%10s-+-%10s-+-%10s\n",$dashes,$dashes,$dashes,$dashes,$dashes));
   $oo->disp(sprintf("%10.10s | %10s | %10s\n","Total",$qualct,$cum_percent));
   $oo->disp("\n");
  } else {
   $r->{status} = "fail; unknown variable $vars[0]";
  }
  return $r;
 } else {
  return { status => "fail; no variable specified" };
 }
}

sub csgof {
 my $self = shift;
 my $a    = shift;

 if (!($a->{firstvar}{present} && ($a->{firstvar}{state} eq "exist"))) {
  return { status => "fail; variable not present" };
 }

 my $v = $a->{firstvar}{name};

 my $missing = tstFlg($a,"missing");
 my $flgstr = "";
 for my $f (@{$a->{flgs}}) {
  $flgstr .= $f . " ";
 }
 my @epf = ();
 my $have_exp = 0;
 if ($flgstr =~ /^expperc\((.*)\)/) {
  $have_exp = 1;
  my $expperc_args = $1;
  my @epcts = split(/\s+/,$expperc_args);
  my $totpct = 0;
  for (my $i=0;$i<@epcts;$i++) {
   if (!looks_like_number($epcts[$i])) {
    return { status => "fail; non-numeric argument to expperc" }
   }
   $epf[$i] = $epcts[$i] / 100;
   $totpct += $epcts[$i];
  }
  if ($totpct > 100) {
    return { status => "fail; you allocated more than 100% of distribution" }
  }
  if ($totpct < 100) {
   push(@epf, (100-$totpct)/100);
  }
 }

 my $q = {};
 my $totct = 0;

 my @vlist = ($v);
 my $qdm;
 if ($missing) {
  $qdm = gen_ifquald_ha($self->{dm},\@vlist,$a->{ifary});
 } else {
  $qdm = gen_defquald_ha($self->{dm},\@vlist,$a->{ifary});
 }

 for (my $y=0;$y<@{$qdm->{data}{$v}};$y++) {
  my $d = $qdm->{data}{$v}[$y];
  if (!defined($q->{$d})) {
   $q->{$d}{actual} = 1;
  } else {
   $q->{$d}{actual} += 1;
  }
  $totct += 1;
 }

 my $i=0;
 my $chi2 = 0;
 my $os = "\n        Chi-square Goodness of Fit\n\n";
 my $dof = keys %{$q};
 $dof -= 1;

 $os .= sprintf("        %s %s %s \n",clip7($v),clip7("expfreq"),clip7("obsfreq"));
 $os .= sprintf("        %s %s %s \n",clip7("-------"),clip7("-------"),clip7("-------"));
 foreach my $n (sort keys %{$q}) {

  if ($have_exp) {
   if (defined($epf[$i])) {
    $q->{$n}{expected} = $epf[$i] * $totct;
   } else {
    return { status => "fail; insufficient number of expected values for categories" }
   }
  } else {
   $q->{$n}{expected} = $totct / ($dof+1);
  }

  my $oe2e = (($q->{$n}{actual} - $q->{$n}{expected})**2) / $q->{$n}{expected};
  $q->{$n}{oe2e} = $oe2e;
  $chi2 += $oe2e;

  $os .= sprintf("        %s       %s %s\n",clip7($n),clip7($q->{$n}{expected}),clip7($q->{$n}{actual}));
  $i++;
 }

 my $oo = $self->{oo};
 $oo->disp($os);

 my $chiprob = Statistics::Distributions::chisqrprob($dof,$chi2);
 my $r = {};
 $oo->disp("\n  chisq($dof) = " . fixedFloat($chi2) . ", p=$chiprob\n\n");
 $r->{chi2}   = $chi2;
 $r->{P_chi2} = $chiprob;
 $r->{dof}    = $dof;
 $r->{cat}    = $q;
 $r->{status} = "ok";
 return $r;
}



1;

